# Copyright (c) 2024 Alethea Katherine Flowers.
# Published under the standard MIT License.
# Full text available at: https://opensource.org/licenses/MIT

from typing import override

import loglady
from loglady import Destination, Record

from .utils import assert_dict_subset


class StubDestination(Destination):
    def __init__(self):
        super().__init__()
        self.records = []

    @override
    def __call__(self, record: Record):
        self.records.append(record)


def test_configure_and_magics():
    dest = StubDestination()
    mgr = loglady.configure(destinations=[dest])

    loglady.info("hello, world!")

    l1 = loglady.bind(context=42)
    l1.warning("Eek!", more_stuff="...")

    mgr.flush()

    assert len(dest.records) == 2
    assert_dict_subset(
        dest.records[0],
        dict(
            msg="hello, world!",
            level="info",
            timestamp=...,
            call_filename=...,
            call_fn=...,
            call_lineno=...,
            call_module=...,
            thread_id=...,
            thread_name=...,
        ),
    )
    assert_dict_subset(
        dest.records[1],
        dict(
            msg="Eek!",
            level="warning",
            more_stuff="...",
            timestamp=...,
            call_filename=...,
            call_fn=...,
            call_lineno=...,
            call_module=...,
            thread_id=...,
            thread_name=...,
        ),
    )



# Copyright (c) 2024 Alethea Katherine Flowers.
# Published under the standard MIT License.
# Full text available at: https://opensource.org/licenses/MIT

import datetime
import sys
import threading
from types import FrameType

from ._tracebackhide import check_for_tracebackhide
from .types import Record


def add_timestamp(record: Record) -> Record:
    """Adds the current timestamp"""
    record["timestamp"] = datetime.datetime.now()  # noqa: DTZ005
    return record


def add_thread_info(record: Record) -> Record:
    """Adds teh current thread native id and name"""
    current_thread = threading.current_thread()
    record["thread_id"] = current_thread.ident
    record["thread_name"] = current_thread.name
    return record


def add_exception_and_stack_info(record: Record) -> Record:
    """Adds the current exception info and stacktrace

    The exception info is added if record["exc_info"] is True, likewise,
    the stacktrace is added if record["stack_info"] is True.
    """
    if record.pop("exc_info", False):
        record["exception"] = sys.exc_info()

    if record.pop("stack_info", False):
        record["stacktrace"] = _find_app_frame()

    return record


def add_call_info(record: Record) -> Record:
    """Add the calling function's name, filename, module, and lineno"""
    frame = _find_app_frame()
    record["call_filename"] = frame.f_code.co_filename
    record["call_module"] = frame.f_globals["__name__"]
    record["call_fn"] = frame.f_code.co_qualname
    record["call_lineno"] = frame.f_lineno
    return record


def _find_app_frame(stack: FrameType | None = None, ignores=("loglady.")) -> FrameType:
    """Finds the first frame that isn't part of the logging code"""
    if stack is None:
        # sys._getframe is faster than inspect.currentframe()
        stack = sys._getframe(1)  # pyright: ignore[reportPrivateUsage]

    f = stack
    name = f.f_globals.get("__name__") or "?"

    while True:
        traceback_hide = check_for_tracebackhide(f)
        ignore_hide = name.startswith(ignores)

        if not (traceback_hide or ignore_hide):
            break

        if f.f_back is None:
            name = "?"
            break

        f = f.f_back
        name = f.f_globals.get("__name__") or "?"

    return f


def fancy_prefix_icon(record: Record) -> Record:
    icon = record.get("icon", None)

    match icon:
        case ">":
            icon = "➤"
        case "->":
            icon = "🡲"
        case "<-":
            icon = "🡰"
        case "o":
            icon = "●"
        case "...":
            icon = "…"
        case "v":
            icon = "✓"
        case "x":
            icon = "✗"
        case "*":
            icon = "🟊"
        case "**":
            icon = "🞷"
        case "+":
            icon = "✦"
        case "s":
            icon = "§"
        case "p":
            icon = "¶"
        case "!!":
            icon = "‼︎"
        case "!?":
            icon = "⁉︎"
        case "?!":
            icon = "⁈"
        case "??":
            icon = "⁇"
        case "<3":
            icon = "❤︎"
        case ":)":
            icon = "☺︎"
        case ":(":
            icon = "☹︎"
        case "f":
            icon = "⚑"
        case "snow":
            icon = "☃︎"
        case _:
            pass

    if icon:
        record["icon"] = icon

    return record



# Copyright (c) 2024 Alethea Katherine Flowers.
# Published under the standard MIT License.
# Full text available at: https://opensource.org/licenses/MIT


def assert_dict_subset(actual, expected):
    """Compares two dictionaries but only checks the keys found in *expected*.

    Also allows specifying ... as the value in the expected dictionary to accept any value in the actual as long as the
    key is present.
    """
    __tracebackhide__ = True

    for k, rv in expected.items():
        try:
            lv = actual[k]
        except KeyError as err:
            msg = f'"{k}" not found in actual, expected actual["{k}"]=={rv!r}'
            raise AssertionError() from err

        if rv == ...:
            continue

        if lv != rv:
            msg = f'expected actual["{k}"]=={rv!r}, found {lv!r}'
            raise AssertionError(msg)



# Copyright (c) 2024 Alethea Katherine Flowers.
# Published under the standard MIT License.
# Full text available at: https://opensource.org/licenses/MIT

"""Magics are sugary convenience methods to quickly and easily log to your
heart's content.

These are exposed as top-level methods on the loglady module:

    import loglady

    loglady.configure()

    loglady.info("Hello!")

Magics all work as long as configure() has been called. They match the methods
found on Logger.
"""

from typing import Any

from . import _config_stack
from .logger import Logger


def _cached_logger() -> Logger:
    state = _config_stack.top()

    if not state:
        msg = "You must call loglady.configure() before using the magic log methods."
        raise RuntimeError(msg)

    if state.logger is None:
        state.logger = state.manager.logger()

    return state.logger


def log(msg, **record: Any) -> None:
    _cached_logger().log(msg, **record)


def trace(msg, **record: Any) -> None:
    return _cached_logger().trace(msg, **record)


def debug(msg, **record: Any) -> None:
    _cached_logger().debug(msg, **record)


def warning(msg, **record: Any) -> None:
    return _cached_logger().warning(msg, **record)


warn = warning


def info(msg, **record: Any) -> None:
    _cached_logger().info(msg, **record)


def success(msg, **record: Any) -> None:
    _cached_logger().success(msg, **record)


def error(msg, **record: Any) -> None:
    _cached_logger().error(msg, **record)


def exception(msg, **record: Any) -> None:
    _cached_logger().exception(msg, **record)


def bind(**context: Any) -> Logger:
    return _cached_logger().bind(**context)


def prefix(prefix: str, **context: Any) -> Logger:
    return _cached_logger().prefix(prefix, **context)


def catch(exc_types=BaseException, *, msg: str = "unexpected error", reraise: bool = False):
    return _cached_logger().catch(exc_types=exc_types, msg=msg, reraise=reraise)


def flush():
    state = _config_stack.top()

    if not state:
        return

    state.manager.flush()



# Copyright (c) 2024 Alethea Katherine Flowers.
# Published under the standard MIT License.
# Full text available at: https://opensource.org/licenses/MIT

"""Demo for LogLady

Run using:
    python3 -m loglady
"""

import datetime
from decimal import Decimal

import loglady

_counter = 0


def add_mock_timestamp(record: loglady.Record) -> loglady.Record:
    global _counter  # noqa: PLW0603
    record["timestamp"] = datetime.datetime(  # noqa: DTZ001
        year=2024,
        month=3,
        day=15,
        hour=0,
        minute=round(_counter),
        second=0,
    )
    _counter += 0.25
    return record


def demo_prefixes(log: loglady.Logger):
    l2 = log.prefix("prefix")
    l2.debug("there's a", icon=">")
    l2.info("bunch of neat", icon="->")
    l2.warning("icons that can be!", icon="<-")
    l2.success("used with prefix!", icon="o")
    l2.error("including arrows, marks, and other fun bits!", icon="...")
    l2.debug("there's a", icon="v")
    l2.info("bunch of neat", icon="x")
    l2.warning("icons that can be!", icon="*")
    l2.success("used with prefix!", icon="**")
    l2.error("including arrows, marks, and other fun bits!", icon="+")
    l2.debug("there's a", icon="s")
    l2.info("bunch of neat", icon="p")
    l2.warning("icons that can be!", icon="!!")
    l2.success("used with prefix!", icon="??")
    l2.error("including arrows, marks, and other fun bits!", icon="?!")
    l2.debug("there's a", icon="<3")
    l2.info("bunch of neat", icon=":)")
    l2.warning("icons that can be!", icon=":(")
    l2.success("used with prefix!", icon="f")
    l2.error("including arrows, marks, and other fun bits!", icon="snow")


def demo_magics():
    loglady.debug("via magic loglady.debug")
    loglady.info("via magic loglady.info")
    loglady.warning("via magic loglady.warning")
    loglady.success("via magic loglady.success")
    loglady.error("via magic loglady.error")


def demo_exc_and_stack(log: loglady.Logger):
    try:
        lol_this_wont_work()  # pyright: ignore[reportUndefinedVariable]  # noqa: F821
    except Exception:
        log.exception("this one has an exception attached")

    log.trace("& this one has a stacktrace!")


def demo_catcher(log: loglady.Logger):
    with log.catch(msg="this catches exceptions with a context manager"):
        lol_this_wont_work()  # pyright: ignore[reportUndefinedVariable]  # noqa: F821


class DemoCallsite:
    def __call__(self):
        def inner():
            log.info("this log message is nestled deep!")

        inner()


if __name__ == "__main__":
    mgr = loglady.configure(
        middleware=[*loglady.DEFAULT_MIDDLEWARE, add_mock_timestamp],
    )

    log = mgr.logger()
    log.debug("this is a debug message")
    log.info("this one is an info message")
    log.warning("watch out, this one is a warning!")
    log.success("oh nice, this one is a success!")
    log.error("oops, this one is an error")

    log.info(
        "This one has structured data!",
        the_answer="42",
        thing=dict(key="value"),
        decimal=Decimal("3.14"),
        a_class=loglady.Logger,
    )

    log.info(
        """\
this one has a really really really really really long message that will get split over \
multiple lines. You are a worm through time. The thunder song distorts you. Happiness \
comes. White pearls, but yellow and red in the eye. Through a mirror, inverted is made \
right. Leave your insides by the door. Push the fingers through the surface into the \
wet. You've always been the new you. You want this to be true. We stand around you while \
you dream. You can almost hear our words but you forget. This happens more and more now. \
You gave us the permission in your regulations. We wait in the stains. The word that \
describes this is redacted. Repeat the word.""",
        it="also has",
        some=dict(data=42),
    )

    demo_magics()
    DemoCallsite()()
    demo_prefixes(log)
    demo_exc_and_stack(log)
    demo_catcher(log)



# Copyright (c) 2024 Alethea Katherine Flowers.
# Published under the standard MIT License.
# Full text available at: https://opensource.org/licenses/MIT

from .destination import RichConsoleDestination

__all__ = ["RichConsoleDestination"]



# Copyright (c) 2024 Alethea Katherine Flowers.
# Published under the standard MIT License.
# Full text available at: https://opensource.org/licenses/MIT


from types import FrameType

import loglady.middleware
from loglady.types import SysExcInfo


def test_add_stack_info():
    record = loglady.middleware.add_exception_and_stack_info(dict(stack_info=True))

    stack: FrameType | None = record.get("stacktrace")
    assert stack is not None
    assert stack.f_globals.get("__file__") == __file__


def test_add_exception_info():
    record = dict(exc_info=True)
    err = RuntimeError()

    try:
        raise err
    except RuntimeError:
        record = loglady.middleware.add_exception_and_stack_info(record)

    excinfo: SysExcInfo | None = record.get("exception")
    assert excinfo is not None

    exc_type, exc, traceback = excinfo

    assert exc_type == RuntimeError
    assert exc == err
    assert traceback is not None


def test_add_call_info():
    record = loglady.middleware.add_call_info(dict())

    assert record.get("call_fn") == "test_add_call_info"
    assert record.get("call_filename") == __file__
    assert record.get("call_module") == __name__


def test_add_call_info_with_invisible_fn():
    def invisible_fn():
        __tracebackhide__ = True
        return loglady.middleware.add_call_info(dict())

    record = invisible_fn()

    assert record.get("call_fn") == "test_add_call_info_with_invisible_fn"
    assert record.get("call_filename") == __file__
    assert record.get("call_module") == __name__



# Copyright (c) 2024 Alethea Katherine Flowers.
# Published under the standard MIT License.
# Full text available at: https://opensource.org/licenses/MIT


import contextlib
from types import FrameType


def check_for_tracebackhide(frame: FrameType) -> bool:
    """Helper for ignoring stackframes from functions that have a variable
    named __stacktracehide__ defined.

    This is adapted from pytest's TracebackEntry.ishidden
    """
    tbh = False

    # Note that f_locals and f_globals are normally dictionaries, but with
    # exec() and eval() they can be set to just about anything, so we swallow
    # all exceptions while trying this nonsense.
    with contextlib.suppress(Exception):
        for namespace in (frame.f_locals, frame.f_globals):
            tbh = namespace.get("__tracebackhide__", None)
            if tbh is not None:
                break

    if tbh and callable(tbh):
        tbh = tbh(frame)

    return tbh



# Copyright (c) 2024 Alethea Katherine Flowers.
# Published under the standard MIT License.
# Full text available at: https://opensource.org/licenses/MIT

"""
Formatters for RichConsoleDestination
"""

from collections.abc import Callable, Iterable
from dataclasses import dataclass
from types import MappingProxyType, ModuleType

import rich
import rich.box
import rich.console
import rich.containers
import rich.highlighter
import rich.pretty
import rich.table
import rich.traceback
from rich.text import Text

from loglady.threading import thread_emoji
from loglady.types import Record

from ._stacktrace import Stacktrace

type Formatter = Callable[[Record], rich.console.RenderableType | None]


@dataclass
class ExceptionFormatter:
    width: int | None = None
    extra_lines: int = 1
    theme: str | None = "github-dark"
    word_wrap: bool = False
    show_locals: bool = True
    locals_max_length: int = 10
    locals_max_string: int = 80
    locals_hide_dunder: bool = True
    locals_hide_sunder: bool = False
    indent_guides: bool = True
    suppress: Iterable[str | ModuleType] = ()
    max_frames: int = 100

    def __call__(self, record):
        exc = record.pop("exception", None)

        if exc is None or exc == (None, None, None):
            return None

        return rich.traceback.Traceback.from_exception(
            *exc,
            width=self.width,
            extra_lines=self.extra_lines,
            theme=self.theme,
            word_wrap=self.word_wrap,
            show_locals=self.show_locals,
            locals_max_length=self.locals_max_length,
            locals_max_string=self.locals_max_string,
            locals_hide_dunder=self.locals_hide_dunder,
            locals_hide_sunder=self.locals_hide_sunder,
            indent_guides=self.indent_guides,
            suppress=self.suppress,
            max_frames=self.max_frames,
        )


class StacktraceFormatter:
    def __init__(self):
        super().__init__()

    def __call__(self, record):
        if (stack := record.pop("stacktrace", None)) is None:
            return None
        return Stacktrace(stack)


DEFAULT_LEVEL_TO_TEXT = MappingProxyType(
    dict(
        error=Text("█", style="log.level.error"),
        warning=Text("█", style="log.level.warning"),
        success=Text("█", style="log.level.success"),
        info=Text("█", style="log.level.info"),
        debug=Text("█", style="log.level.debug"),
        notset=Text("█", style="log.level.notset"),
    )
)


class LevelFormatter:
    def __init__(self, level_to_text=DEFAULT_LEVEL_TO_TEXT):
        super().__init__()
        self.level_to_text = level_to_text

    def __call__(self, record: Record):
        level = record.get("level", "notset")
        sash = self.level_to_text.get(level)
        return sash


class MessageFormatter:
    def __call__(self, record: Record):
        level = record.get("level", "notset")
        msg = record.pop("msg")
        prefix = record.pop("prefix", None)
        icon = record.pop("icon", "●" if prefix else "")
        if icon:
            icon = f" {icon} "

        formatted = Text.from_markup(text=f"{prefix if prefix else ""}{icon}{msg} ", style=f"log.level.{level}")

        return formatted


class TimestampFormatter:
    def __call__(self, record: Record):
        timestamp = record.pop("timestamp", None)

        if not timestamp:
            return None

        return timestamp.strftime(format="%H:%M")


class CallsiteFormatter:
    def __init__(self, *, include_module: bool = False, collapse_special: bool = True):
        super().__init__()
        self.include_module = include_module
        self.collapse_special = collapse_special

    def __call__(self, record: Record):
        func_name = record.pop("call_fn", None)

        if not func_name:
            return "."

        filename = record.pop("call_filename")
        module = record.pop("call_module")
        lineno = record.pop("call_lineno")

        if self.include_module:
            name = f"{module}:{func_name}()"
        else:
            name = f"{func_name}()"

        if self.collapse_special:
            name = name.replace("<module>", f"{module}()")
            name = name.replace(".<locals>", "()")
            name = name.replace(".__call__", "()")
            name = name.replace("()()", "()")

        return Text(name, style=f"link file://{filename}:{lineno}")


class NonrepeatedFormatter:
    def __init__(self, fn: Formatter, *, fillchar: str = "⋅", fill: bool = False):
        super().__init__()
        self._fn = fn
        self._fillchar = fillchar
        self._fill = fill
        self._last = None

    def __call__(self, record: Record):
        new = self._fn(record)

        if new == self._last:
            if self._fill:
                length = 1
                match new:
                    case str() | Text():
                        length = len(new)
                    case _:
                        length = 1
                new = Text(self._fillchar * length, style="log.repeated")
            else:
                new = Text(self._fillchar, style="log.repeated")
        else:
            self._last = new

        return new


DEFAULT_IGNORED_KEYS = frozenset(
    {
        "level",
        "msg",
        "prefix",
        "icon",
        "timestamp",
        "call_fn",
        "call_filename",
        "call_module",
        "call_lineno",
        "thread_id",
        "thread_name",
        "exception",
        "stacktrace",
    }
)


class RecordItemsFormatter:
    def __init__(self, ignored_keys=DEFAULT_IGNORED_KEYS):
        super().__init__()
        self.ignored_keys = ignored_keys
        self._hl = rich.highlighter.ReprHighlighter()

    def __call__(self, record: Record):
        return Text.assemble(*self._gen_items(record))

    def _gen_items(self, record: Record):
        for k, v in record.items():
            if k in self.ignored_keys:
                continue

            yield Text(f"{k}=", "log.items.keys")
            match v:
                case bool():
                    yield Text("true", "repr.bool_true") if v else Text("false", "repr.bool_false")
                case None:
                    yield Text("none", "repr.none")
                case _:
                    yield self._hl(repr(v))
            yield " "


class ThreadInfoFormatter:
    def __call__(self, record: Record):
        id_ = record.pop("thread_id", 0)
        _ = record.pop("thread_name", "")
        emoji = thread_emoji(id_)
        return Text(text=emoji)



# Copyright (c) 2024 Alethea Katherine Flowers.
# Published under the standard MIT License.
# Full text available at: https://opensource.org/licenses/MIT

import queue
import threading
from abc import ABC, abstractmethod
from typing import override
from warnings import warn

from .destination import DestinationList
from .types import Record


class Transport(ABC):
    """Transports are responsible for relaying records to a list of destinations"""

    destinations: DestinationList = ()

    def __init__(self):
        super().__init__()

    @abstractmethod
    def relay(self, record: Record) -> None:
        raise NotImplementedError()

    def process(self, record: Record) -> None:
        for dest in self.destinations:
            dest(record)

    def start(self) -> None:
        return

    def stop(self) -> None:
        return

    def flush(self) -> None:
        for dest in self.destinations:
            dest.flush()


class SyncTransport(Transport):
    """A very simple transport that immediately relays records to destinations.

    This is useful for tests but not recommended for real applications.
    """

    @override
    def relay(self, record: Record) -> None:
        self.process(record)


class ThreadedTransport(Transport):
    """A transport that handles relaying in a separate thread.

    This prevents logging calls from blocking, as they only need to queue the
    record.
    """

    _STOP = object()
    _FLUSH = object()

    def __init__(self):
        super().__init__()
        self._q = queue.SimpleQueue()
        self._thread: threading.Thread | None = None
        self._flush_cond = threading.Condition()

    @override
    def relay(self, record: Record):
        self._q.put(record)

    def _thread_main(self):
        while True:
            try:
                record = self._q.get(block=True)
                if record is self._STOP:
                    break

                if record is self._FLUSH:
                    with self._flush_cond:
                        self._flush_cond.notify_all()
                    continue

                self.process(record)

            except queue.Empty:
                break

    @override
    def start(self):
        self._thread = threading.Thread(target=self._thread_main)
        self._thread.daemon = True
        self._thread.start()

    @override
    def stop(self):
        if self._thread is None:
            return

        self._q.put(self._STOP)
        self._thread.join()
        self._thread = None

        if not self._q.empty():
            warn(
                f"logging thread exited with {self._q.qsize()} logs not yet processed",
                stacklevel=1,
            )

    @override
    def flush(self):
        if self._thread is None:
            return

        self._q.put(self._FLUSH)

        with self._flush_cond:
            _ = self._flush_cond.wait()

        super().flush()



# Copyright (c) 2024 Alethea Katherine Flowers.
# Published under the standard MIT License.
# Full text available at: https://opensource.org/licenses/MIT

from .destination import DestinationList
from .logger import Logger
from .transport import Transport
from .types import MiddlewareList, Record


class Manager:
    """Wrangles all the bits of LogLady so that things work!

    Managers can be constructed directly, though typically you'll use
    loglady.configure() so that it gets globally registered as the active
    manager.

    A manager gives a Logger life: Any records created by a Logger are sent to
    the manager. The manager runs the record through middleware and then hands
    it off to the transport. The transport is responsible for sending the record
    to all destinations.

    Managers are responsible for the lifetime of its Transport and Destinations.
    If you're manually creating managers, don't forget to call start() and
    stop().
    """

    def __init__(
        self,
        *,
        transport: Transport,
        middleware: MiddlewareList,
        destinations: DestinationList,
    ):
        super().__init__()
        self.transport = transport
        self.middleware = middleware
        self.destinations = destinations

    @property
    def destinations(self) -> DestinationList:
        return self.transport.destinations

    @destinations.setter
    def destinations(self, val: DestinationList):
        self.transport.destinations = val

    def logger(self, **context):
        """Get a new Logger"""
        return Logger(relay=self.relay, context=context)

    def start(self):
        """Tell the transport to start"""
        self.transport.start()

    def stop(self):
        """Flush all destinations and stop the transport"""
        self.flush()
        self.transport.stop()

    def flush(self):
        """Ask all destinations to write any pending logs"""
        self.transport.flush()

    def _apply_middleware(self, record: Record | None):
        for fn in self.middleware:
            if record is None:
                break
            record = fn(record)

        return record

    def relay(self, record: Record) -> None:
        """Run a record through middleware and hand it off to the transport"""
        processed_record = self._apply_middleware(record)
        if processed_record is None:
            return
        self.transport.relay(processed_record)



# Copyright (c) 2024 Alethea Katherine Flowers.
# Published under the standard MIT License.
# Full text available at: https://opensource.org/licenses/MIT

import loglady

from .utils import assert_dict_subset


def test_capture(loglady_capture):
    loglady.info("Hello!")
    assert len(loglady_capture) == 1
    assert_dict_subset(loglady_capture[0], dict(msg="Hello!"))


def test_stdout(capsys):
    loglady.info("Hello!")
    captured = capsys.readouterr()
    # Loglady should *not* be writing to stdout be default during tests
    assert "Hello!" not in captured.out



# Copyright (c) 2024 Alethea Katherine Flowers.
# Published under the standard MIT License.
# Full text available at: https://opensource.org/licenses/MIT

"""Global configuration for LogLady.

This global config is used by magics (the top-level loglady.info, etc.), and
should be configured at application startup.
"""

import sys

from . import _config_stack
from .destination import DestinationList
from .manager import Manager
from .middleware import add_call_info, add_exception_and_stack_info, add_thread_info, add_timestamp, fancy_prefix_icon
from .rich import RichConsoleDestination
from .transport import SyncTransport, ThreadedTransport, Transport
from .types import MiddlewareList

DEFAULT_MIDDLEWARE = (
    add_timestamp,
    add_thread_info,
    add_exception_and_stack_info,
    add_call_info,
    fancy_prefix_icon,
)


def configure(
    *,
    transport: Transport | None = None,
    middleware: MiddlewareList = DEFAULT_MIDDLEWARE,
    destinations: DestinationList | None = None,
    once: bool = False,
    install_hook: bool = True,
) -> Manager:
    """Configure LogLady.

    This creates global configuration for LogLady that's then available via the
    top-level loglady.info(...), loglady.logger(...), etc. methods (these are
    called "magics").

    This creates a Manager instance and start()s it so that any background
    stuff can happen. It also installs an atexit() handler to call the Manager's
    stop() to ensure all logs are written before exit.
    """
    if once and _config_stack.top():
        return manager()

    if transport is None:
        if _is_repl():
            transport = SyncTransport()
        else:
            transport = ThreadedTransport()

    if destinations is None:
        destinations = [RichConsoleDestination()]

    mgr = Manager(
        transport=transport,
        middleware=middleware,
        destinations=destinations,
    )
    mgr.start()

    _config_stack.push(_config_stack.GlobalState(mgr))

    if install_hook and sys.excepthook == sys.__excepthook__:
        sys.excepthook = _excepthook

    return mgr


def manager() -> Manager:
    """Returns the currently configured Manager, or None if not configured."""
    state = _config_stack.top()
    if state is None:
        msg = "LogLady has not been configured!"
        raise RuntimeError(msg)
    return state.manager


def _excepthook(exc_type, value, traceback):
    state = _config_stack.top()
    if state is None:
        sys.__excepthook__(exc_type, value, traceback)
        return

    state.manager.flush()

    state.manager.logger().log(
        "unhandled exception, sys.excepthook() called",
        level="error",
        exception=(
            exc_type,
            value,
            traceback,
        ),
    )

    state.manager.flush()
    state.manager.stop()


def _is_repl() -> bool:
    return hasattr(sys, "ps1")



# Copyright (c) 2024 Alethea Katherine Flowers.
# Published under the standard MIT License.
# Full text available at: https://opensource.org/licenses/MIT

from typing import override

from loglady import Destination, Record
from loglady.transport import SyncTransport, ThreadedTransport


class StubDestination(Destination):
    def __init__(self):
        super().__init__()
        self.records = []

    @override
    def __call__(self, record: Record):
        self.records.append(record)


def test_sync_transport():
    transp = SyncTransport()
    record = dict(a=42, b="hello!")

    # No destinations, shouldn't do anything but also shouldn't error.
    transp.relay(record)

    dest = StubDestination()
    transp.destinations = [dest]

    transp.relay(record)

    assert dest.records.pop() == dict(a=42, b="hello!")


def test_threaded_transport():
    transp = ThreadedTransport()
    dest = StubDestination()
    transp.destinations = [dest]

    record = dict(a=42, b="hello!")

    transp.relay(record)

    # Not yet processed, so it shouldn't be in the destination.
    assert len(dest.records) == 0

    # Start the thread, flush, and check again.
    transp.start()
    transp.flush()

    assert dest.records.pop() == dict(a=42, b="hello!")

    transp.stop()



# Copyright (c) 2024 Alethea Katherine Flowers.
# Published under the standard MIT License.
# Full text available at: https://opensource.org/licenses/MIT

from loglady import Record
from loglady.logger import Logger


class RelayStub:
    def __init__(self):
        super().__init__()
        self.records = []

    def __call__(self, record: Record) -> None:
        self.records.append(record)


def test_construct():
    relay = RelayStub()

    logger = Logger(relay=relay)
    assert logger.context == {}

    logger = Logger(relay=relay, context=dict(a=42))
    assert logger.context == dict(a=42)


def test_bind_unbind():
    relay = RelayStub()

    l_root = Logger(relay=relay)
    l1 = l_root.bind(a=42, b="two")
    l2 = l1.bind(a=43, c="three")
    l3 = l1.bind(a=None, d="four")

    assert l_root.context == {}
    assert l1.context == dict(a=42, b="two")
    assert l2.context == dict(a=43, b="two", c="three")
    assert l3.context == dict(a=None, b="two", d="four")

    l4 = l1.unbind("b", "d")
    assert l4.context == dict(a=42)
    assert l1.context != l4.context


def test_methods():
    relay = RelayStub()
    log = Logger(relay=relay)

    log.log("hello", a=42)
    assert relay.records.pop() == dict(msg="hello", a=42)
    log.debug("hello", a=43)
    assert relay.records.pop() == dict(msg="hello", level="debug", a=43)
    log.trace("hello", a=44)
    assert relay.records.pop() == dict(msg="hello", level="debug", stack_info=True, a=44)
    log.info("hello", a=45)
    assert relay.records.pop() == dict(msg="hello", level="info", a=45)
    log.warning("hello", a=46)
    assert relay.records.pop() == dict(msg="hello", level="warning", a=46)
    log.warn("hello", a=47)
    assert relay.records.pop() == dict(msg="hello", level="warning", a=47)
    log.success("hello", a=48)
    assert relay.records.pop() == dict(msg="hello", level="success", a=48)
    log.error("hello", a=49)
    assert relay.records.pop() == dict(msg="hello", level="error", a=49)
    log.exception("hello", a=50)
    assert relay.records.pop() == dict(msg="hello", level="error", exc_info=True, a=50)


def test_methods_with_context():
    relay = RelayStub()
    log = Logger(relay=relay).bind(context_a=42)

    log.log("hello")
    assert relay.records.pop() == dict(msg="hello", context_a=42)



# Copyright (c) 2024 Alethea Katherine Flowers.
# Published under the standard MIT License.
# Full text available at: https://opensource.org/licenses/MIT

from .config import DEFAULT_MIDDLEWARE, configure
from .destination import CaptureDestination, Destination, TextIODestination
from .logger import Logger
from .magics import bind, catch, debug, error, exception, flush, info, log, success, trace, warn, warning
from .manager import Manager
from .middleware import add_call_info, add_exception_and_stack_info, add_thread_info, add_timestamp, fancy_prefix_icon
from .rich import RichConsoleDestination
from .transport import SyncTransport, ThreadedTransport, Transport
from .types import Middleware, Record

__all__ = [
    # Types & classes
    "CaptureDestination",
    "Destination",
    "Logger",
    "Manager",
    "Middleware",
    "Record",
    "RichConsoleDestination",
    "SyncTransport",
    "TextIODestination",
    "ThreadedTransport",
    "Transport",
    # Middleware
    "add_call_info",
    "add_exception_and_stack_info",
    "add_thread_info",
    "add_timestamp",
    "fancy_prefix_icon",
    "DEFAULT_MIDDLEWARE",
    # Magics
    "bind",
    "catch",
    "configure",
    "debug",
    "error",
    "exception",
    "flush",
    "info",
    "log",
    "success",
    "trace",
    "warn",
    "warning",
]



# Copyright (c) 2024 Alethea Katherine Flowers.
# Published under the standard MIT License.
# Full text available at: https://opensource.org/licenses/MIT

"""Global state stack management.

LogLady's magics (the top-level loglady.info, etc.) need to have access to the
manager created from calling configure(), this manages keeping track of that
global state. It's done as a stack to make life easier for tests, as this allows
multiple calls to configure() to not stomp on each other.
"""

import atexit
import contextlib
from dataclasses import dataclass

from .logger import Logger
from .manager import Manager


@dataclass
class GlobalState:
    manager: Manager
    """The manager created via configure() or configure_for_tests()"""

    logger: Logger | None = None
    """The root logger used by magics"""


_STATE_STACK = []


def top() -> GlobalState | None:
    """Returns the currently active global config or None if LogLady hasn't
    been configured."""
    if not _STATE_STACK:
        return None
    return _STATE_STACK[-1]


def push(state: GlobalState):
    """Add the given state to the top of the stack, making it active."""
    return _STATE_STACK.append(state)


def pop() -> GlobalState | None:
    """Remove the currently active global config from the stack and return it.

    Returns None if the stack is empty."""
    if not _STATE_STACK:
        return None
    return _STATE_STACK.pop()


@contextlib.contextmanager
def rewind():
    """A context manager that automatically rewinds the stack on exit.

    This is useful for applying temporary configuration in tests and such.
    """
    checkpoint_size = len(_STATE_STACK)
    try:
        yield
    finally:
        assert len(_STATE_STACK) >= checkpoint_size
        while len(_STATE_STACK) != checkpoint_size:
            state = pop()
            if state is not None:
                state.manager.stop()


def _on_shutdown():
    while state := pop():
        state.manager.stop()


_ = atexit.register(_on_shutdown)



# Copyright (c) 2024 Alethea Katherine Flowers.
# Published under the standard MIT License.
# Full text available at: https://opensource.org/licenses/MIT

"""Rich-compatible stacktrace printer.

Rich doesn't have a way to nicely print stacktraces without an Exception involved,
so this module allows for log.trace() to print out the stacktrace.

It's a lot of code. It's mostly because Thea can't leave well enough alone.
"""

import linecache
import os
from collections.abc import Iterator
from types import FrameType
from typing import Any

import pygments.token
from rich import box
from rich.console import Console, ConsoleOptions, RenderResult, group
from rich.constrain import Constrain
from rich.highlighter import RegexHighlighter
from rich.panel import Panel
from rich.pretty import Pretty
from rich.style import Style
from rich.syntax import Syntax
from rich.table import Table
from rich.text import Text
from rich.theme import Theme

from loglady._tracebackhide import check_for_tracebackhide


class Stacktrace:
    def __init__(self, stack, *, width: int | None = None, extra_lines: int = 3, syntax_theme: str = "github-dark"):
        super().__init__()
        self.stack = stack
        self.width = width
        self.extra_lines = extra_lines
        self.syntax_theme = syntax_theme

    def __rich_console__(self, console: Console, options: ConsoleOptions) -> RenderResult:
        frames = reversed(list(_iter_stack(self.stack)))

        traceback_theme = _make_theme()

        with console.use_theme(traceback_theme):
            frame_renderable = Panel(
                self._render_frames(frames),
                title="[stacktrace.title]Stacktrace",
                subtitle="[dim](most recent call last)",
                border_style="stacktrace.border",
                expand=True,
                padding=(0, 1),
            )

            stack_renderable = Constrain(frame_renderable, self.width)

            yield stack_renderable

    @group()
    def _render_frames(self, frames: Iterator["_Stackframe"]):
        for n, frame in enumerate(frames):
            yield frame.render(n, extra_lines=self.extra_lines, syntax_theme=self.syntax_theme)


class _Stackframe:
    def __init__(self, frame: FrameType):
        super().__init__()
        self.filename = frame.f_code.co_filename
        self.lineno = frame.f_lineno
        self.fn = frame.f_code.co_name
        self.is_from_file = not self.filename.startswith("<")
        self.is_module = self.fn.startswith("<")
        if self.is_from_file and not self.is_module:
            self.locals = frame.f_locals
        else:
            self.locals = None

        self.is_hidden = (not self.is_from_file) or check_for_tracebackhide(frame)

    @property
    def code(self):
        return "".join(linecache.getlines(self.filename))

    @property
    def relpath(self):
        relpath = os.path.relpath(self.filename)

        if relpath != self.filename:
            relpath = f"./{relpath}"

        return relpath

    @group()
    def render(self, n: int, *, extra_lines: int, syntax_theme: str):
        title = self.title(n=n)

        try:
            body = self.body(extra_lines=extra_lines, syntax_theme=syntax_theme)
        except Exception as err:  # noqa: BLE001
            body = Text(f"\n{err}", "traceback.error")

        if n > 0 and self.is_from_file:
            yield ""

        yield title
        yield body

    def title(self, n: int):
        # Path and function name
        path = self.relpath if self.is_from_file else self.filename
        highlighter = _PathHighlighter()

        return Text.assemble(
            (f"#{n}:", "stacktrace.title"),
            highlighter(Text(path, style="stacktrace.title.filename")),
            (":", "stacktrace.title"),
            (str(self.lineno), "stacktrace.title.lineno"),
            (" @ ", "stacktrace.title"),
            (self.fn, "stacktrace.title.function"),
        )

    @group()
    def body(self, *, extra_lines: int, syntax_theme: str):
        code = self.code
        if not code:
            return

        yield Syntax(
            code,
            "python",
            line_numbers=True,
            line_range=(
                self.lineno - extra_lines,
                self.lineno + extra_lines,
            ),
            highlight_lines={self.lineno},
            word_wrap=False,
            indent_guides=True,
            dedent=False,
            theme=syntax_theme,
        )

        if self.locals:
            yield Panel(
                _render_locals(
                    self.locals,
                    hide_dunder=True,
                    hide_sunder=False,
                ),
                box=box.MINIMAL,
                style="background",
                padding=(0, 0),
            )


class _PathHighlighter(RegexHighlighter):
    def __init__(self):
        super().__init__()
        self.highlights = [r"(?P<inherit>.*/)(?P<bold>.+)"]


def _iter_stack(stack: FrameType):
    frame = stack
    while frame:
        sf = _Stackframe(frame)
        if not sf.is_hidden:
            yield sf
        frame = frame.f_back


def _render_locals(frame_locals: dict[str, Any], *, hide_dunder: bool, hide_sunder: bool):
    """Render a frame's local variables."""
    table = Table.grid(padding=(0, 1), expand=True)
    table.add_column(justify="right")
    table.add_column(justify="left", ratio=1)

    for key, value in frame_locals.items():
        if hide_dunder and key.startswith("__"):
            continue
        if hide_sunder and key.startswith("_"):
            continue

        table.add_row(
            Text.assemble(
                (key, "scope.key.special" if key.startswith("__") else "scope.key"),
                (" =", "scope.equals"),
            ),
            Pretty(value),
        )

    return table


def _make_theme(name: str = "dracula") -> Theme:
    """Make a Rich theme from a pygments theme. Based on code in rich.traceback"""
    syntax_theme = Syntax.get_theme(name)
    token_style = syntax_theme.get_style_for_token
    tks = pygments.token
    traceback_theme = Theme(
        {
            "background": syntax_theme.get_background_style(),
            "pretty": token_style(tks.Text),
            "pygments.text": token_style(tks.Token),
            "pygments.string": token_style(tks.String),
            "pygments.function": token_style(tks.Name.Function),
            "pygments.number": token_style(tks.Number),
            "repr.indent": token_style(tks.Comment) + Style(dim=True),
            "repr.str": token_style(tks.String),
            "repr.brace": token_style(tks.Text) + Style(bold=True),
            "repr.number": token_style(tks.Number),
            "repr.bool_true": token_style(tks.Keyword.Constant),
            "repr.bool_false": token_style(tks.Keyword.Constant),
            "repr.none": token_style(tks.Keyword.Constant),
            "scope.border": token_style(tks.String.Delimiter),
            "scope.equals": token_style(tks.Operator),
            "scope.key": token_style(tks.Name),
            "scope.key.special": token_style(tks.Name.Constant) + Style(dim=True),
        },
        inherit=False,
    )
    return traceback_theme



# Copyright (c) 2024 Alethea Katherine Flowers.
# Published under the standard MIT License.
# Full text available at: https://opensource.org/licenses/MIT

__version__ = "2024.11.29"
__author__ = "Alethea Katherine Flowers"
__author_email__ = "me@thea.codes"



# Copyright (c) 2024 Alethea Katherine Flowers.
# Published under the standard MIT License.
# Full text available at: https://opensource.org/licenses/MIT

from __future__ import annotations

from abc import ABC, abstractmethod
from collections.abc import Callable, Sequence
from functools import cached_property
from io import StringIO
from typing import Protocol, override

from .types import Record


class Destination(ABC):
    """A destination is responsible for *outputting* a Record. They're the last link in the chain."""

    def flush(self):
        """If this destination buffers output, flush it and block until all records have been outputted."""
        return

    @abstractmethod
    def __call__(self, record: Record) -> None:
        """Output the given record to the destination."""
        raise NotImplementedError()


type DestinationList = Sequence[Destination]


type TextIODestinationFormatter = Callable[[Record], str]


class _TextIO(Protocol):
    """The very limited subset of TextIO that TextIODestination depends on."""

    def write(self, s: str, /) -> int: ...
    def flush(self) -> None: ...


class TextIODestination(Destination):
    """A simple destination that just outputs strings to a TextIO instance."""

    def __init__(self, io: _TextIO, formatter: TextIODestinationFormatter | None = None) -> None:
        super().__init__()
        self.io = io

        if formatter is None:
            formatter = PrintFormatter()
        self.formatter = formatter

    @override
    def __call__(self, record: Record) -> None:
        text = self.formatter(record)
        _ = self.io.write(text)

    @override
    def flush(self):
        self.io.flush()


class PrintFormatter:
    """A simple formatter that just uses print() to format a record."""

    def __call__(self, record: Record) -> str:
        out = StringIO()
        print(record, file=out)
        return out.getvalue()


class CaptureDestination(Destination):
    """A simple destination that records all records."""

    def __init__(self):
        super().__init__()
        self.records = []

    @override
    def __call__(self, record: Record) -> None:
        self.records.append(record)

    def reset(self):
        """Clear all captured records"""
        self.records = []

    def playback(self, *destinations: Destination):
        """Playback recorded records to all given destinations"""
        for record in self.records:
            for destination in destinations:
                destination(record)

    def __enter__(self):
        self.reset()
        return self.records

    def __exit__(self, exc_type, exc_value, traceback):
        self.reset()


class LazyDestination(Destination):
    """A destination that wraps another destination, only creating it when
    actually needed."""

    def __init__(self, factory: Callable[[], Destination]):
        super().__init__()
        self._factory = factory

    @cached_property
    def _wrapped(self):
        return self._factory()

    @override
    def __call__(self, record: Record) -> None:
        self._wrapped(record)

    @override
    def flush(self):
        return super().flush()



# Copyright (c) 2024 Alethea Katherine Flowers.
# Published under the standard MIT License.
# Full text available at: https://opensource.org/licenses/MIT

import contextlib
from collections.abc import Mapping
from types import MappingProxyType
from typing import Any, Self

from .types import Context, Relay


class Logger:
    """It's the logger! You know how to log!"""

    __slots__ = ("_relay", "_context")

    def __init__(self, *, relay: Relay, context: Context | None = None):
        """
        Loggers shouldn't be created directly, instead, use loglady.configure()
        and loglady.logger() to get an instance.
        """
        super().__init__()
        self._relay = relay
        self._context = context if context is not None else {}

    @property
    def context(self) -> Mapping[str, Any]:
        """A read-only view of the current context. Use bind() or unbind() to
        change the context."""
        return MappingProxyType(self._context)

    def bind(self, **context: Any) -> Self:
        """Create a new logger with the given context. The new logger inherits
        this logger's context."""
        ctx = self._context.copy()
        ctx.update(**context)
        return self.__class__(relay=self._relay, context=ctx)

    def unbind(self, *keys: str) -> Self:
        """Create a new logger without the given keys in the context."""
        inst = self.bind()
        for key in keys:
            inst._context.pop(key, None)
        return inst

    def prefix(self, prefix: str, **context) -> Self:
        """Shortcut for self.bind(prefix="...", ...)"""
        context["prefix"] = prefix
        return self.bind(**context)

    def log(self, msg, **record: Any) -> None:
        """You probably don't wanna call this, as it's the common log method
        used by info(), warning(), etc. I mean, you can call it, I'm a
        docstring, not a cop."""
        rec = self._context.copy()
        rec.update(**record)
        rec["msg"] = msg

        self._relay(rec)

    def trace(self, msg, **record: Any) -> None:
        """Log a debug message and include a stack trace.

        Note that middleware.add_exception_and_stack_info or a similar
        middleware must be used otherwise it won't actually add the stack info.
        """
        self.log(msg, level="debug", stack_info=True, **record)

    def debug(self, msg, **record: Any) -> None:
        """Log a debug message"""
        self.log(msg, level="debug", **record)

    def info(self, msg, **record: Any) -> None:
        """Log an info message"""
        self.log(msg, level="info", **record)

    def warning(self, msg, **record: Any) -> None:
        """Log a warning message"""
        self.log(msg, level="warning", **record)

    warn = warning

    def success(self, msg, **record: Any) -> None:
        """Log a success message"""
        self.log(msg, level="success", **record)

    def error(self, msg, **record: Any) -> None:
        """Log an error message"""
        self.log(msg, level="error", **record)

    def exception(self, msg, **record: Any) -> None:
        """Log an error message and include the current exception's traceback

        Note that middleware.add_exception_and_stack_info or a similar
        middleware must be used otherwise it won't actually add the exception
        and traceback.
        """
        self.log(msg, level="error", exc_info=True, **record)

    def catch(self, exc_types=BaseException, *, msg: str = "unexpected error", reraise: bool = False):
        @contextlib.contextmanager
        def catcher():
            try:
                yield
            except exc_types:
                self.exception(msg)
                if reraise:
                    raise

        return catcher()



# Copyright (c) 2024 Alethea Katherine Flowers.
# Published under the standard MIT License.
# Full text available at: https://opensource.org/licenses/MIT


import threading

from loglady.threading import thread_emoji


def test_main_thread_implied():
    emoji = thread_emoji()
    assert emoji == "⭐️"


def test_main_thread_explicit():
    thread = threading.main_thread()
    assert thread_emoji(thread) == "⭐️"
    assert thread_emoji(thread.ident) == "⭐️"


def test_other_thread_explicit():
    def thread_main():
        return

    thread = threading.Thread(target=thread_main)
    thread.start()

    assert thread_emoji(thread) != "⭐️"
    assert thread_emoji(thread.ident) != "⭐️"



# Copyright (c) 2024 Alethea Katherine Flowers.
# Published under the standard MIT License.
# Full text available at: https://opensource.org/licenses/MIT

"""
A logging destination based on Rich's fancy-ass console output.
"""

from collections.abc import Mapping
from typing import IO, override
from warnings import warn

import rich
import rich.box
import rich.console
import rich.containers
import rich.highlighter
import rich.pretty
import rich.table
import rich.theme
import rich.traceback
from rich.text import Text

from loglady.destination import Destination
from loglady.types import Record

from . import formatters

DEFAULT_THEME = rich.theme.Theme(
    {
        "log.callsite": "cyan",
        "log.timestamp": "grey62",
        "log.thread": "default",
        "log.level": "default",
        "log.repeated": "grey23",
        "log.level.error": "bold red",
        "log.level.warning": "yellow",
        "log.level.success": "green",
        "log.level.info": "cyan",
        "log.level.debug": "violet",
        "log.level.notset": "gray70",
        "log.items.keys": "italic grey70",
        "log.items.values": "italic",
        "traceback.border": "red",
        "stacktrace.border": "cyan",
        "stacktrace.title": "cyan",
        "stacktrace.title.filename": "magenta",
        "stacktrace.title.lineno": "aquamarine1",
        "stacktrace.title.function": "chartreuse1",
    }
)


def make_default_formatters():
    return dict(
        level=formatters.LevelFormatter(),
        message=formatters.MessageFormatter(),
        timestamp=formatters.NonrepeatedFormatter(formatters.TimestampFormatter(), fill=True),
        callsite=formatters.NonrepeatedFormatter(formatters.CallsiteFormatter()),
        items=formatters.RecordItemsFormatter(),
        thread=formatters.NonrepeatedFormatter(formatters.ThreadInfoFormatter(), fillchar="⋅", fill=True),
        exception=formatters.ExceptionFormatter(),
        stacktrace=formatters.StacktraceFormatter(),
    )


class LineFormatter:
    def __call__(self, formatted):
        msg_and_items = Text.assemble(formatted["message"], formatted["items"])
        msg_container = rich.containers.Renderables([msg_and_items])

        table = self._make_table()
        table.add_row(
            formatted["timestamp"],
            formatted["level"],
            msg_container,
            formatted["callsite"],
            formatted["thread"],
        )

        yield table

        if exception := formatted.pop("exception", None):
            table = self._make_table()
            table.add_row(
                " " * len(formatted["timestamp"]),
                " " * len(formatted["level"]),
                exception,
                "",
                "",
            )
            yield table

        if stack := formatted.pop("stacktrace", None):
            table = self._make_table()
            table.add_row(
                " " * len(formatted["timestamp"]),
                " " * len(formatted["level"]),
                stack,
                "",
                "",
            )
            yield table

    def _make_table(self):
        table = rich.table.Table.grid(padding=(0, 1), expand=True)
        table.add_column(style="log.timestamp")
        table.add_column(style="log.level", width=1, overflow="crop")
        table.add_column(ratio=1, overflow="fold")
        table.add_column(style="log.callsite", justify="right")
        table.add_column(style="log.thread", width=1, overflow="ignore", justify="right")
        return table


class RichConsoleDestination(Destination):
    def __init__(
        self,
        *,
        formatters: Mapping[str, formatters.Formatter] | None = None,
        line_formatter=None,
        theme=DEFAULT_THEME,
        io: IO[str] | None = None,
        console: rich.console.Console | None = None,
    ):
        super().__init__()

        if formatters is None:
            formatters = make_default_formatters()

        self.formatters = formatters

        if not console:
            console = rich.console.Console(theme=theme, file=io)
        else:
            console.push_theme(theme)

        self.console = console

        if line_formatter is None:
            line_formatter = LineFormatter()
        self.line_formatter = line_formatter

    @override
    def __call__(self, record: Record):
        c = self.console

        formatted = {}
        for name, fn in self.formatters.items():
            try:
                formatted[name] = fn(record)
            except Exception as err:  # noqa: BLE001
                warn(f"exception while invoking formatter {fn!r}: {err!r}", stacklevel=1)

        for line in self.line_formatter(formatted):
            c.print(line)



# Copyright (c) 2024 Alethea Katherine Flowers.
# Published under the standard MIT License.
# Full text available at: https://opensource.org/licenses/MIT

import contextlib
from collections.abc import Generator
from io import StringIO

import pytest
import rich

import loglady
import loglady._config_stack
from loglady.destination import CaptureDestination
from loglady.rich.destination import RichConsoleDestination
from loglady.transport import SyncTransport
from loglady.types import Record


@pytest.hookimpl(trylast=True)
def pytest_configure(config: pytest.Config) -> None:
    _ = config.pluginmanager.register(LogladyPlugin(config), "loglady-plugin")


class LogladyPlugin:
    """
    Plugin for pytest that gathers global logs and associates them with tests.

    This is intended to function like pytest's build-in capsys and caplog features.

    Refs:
    - https://github.com/pytest-dev/pytest/blob/72c682ff9773ad2690711105a100423ebf7c7c15/src/_pytest/capture.py#L709-L710
    """

    def __init__(self, config: pytest.Config) -> None:
        super().__init__()
        self.config = config
        self._global_captured = None
        self._fixture_captured = None
        self._manager = None
        self._has_fixture = False

    def start_global_capturing(self):
        self._global_captured = CaptureDestination()
        self._manager = loglady.config.configure(
            transport=SyncTransport(),
            middleware=loglady.config.DEFAULT_MIDDLEWARE,
            destinations=[self._global_captured],
            install_hook=False,
            once=False,
        )

    def stop_global_capturing(self):
        if self._manager is None:
            return

        self._manager.flush()
        self._manager.stop()
        _ = loglady._config_stack.pop()

    def read_global_capture(self):
        if self._global_captured is None:
            return ""

        if not self._global_captured.records:
            return ""

        io = StringIO()
        use_color = self.config.option.color != "no"
        console = rich.console.Console(
            file=io,
            force_terminal=use_color,
            force_interactive=False,
            force_jupyter=False,
        )
        console_dest = RichConsoleDestination(console=console)

        self._global_captured.playback(console_dest)

        return io.getvalue()

    def enable_fixture(self):
        self._fixture_captured = CaptureDestination()
        return self._fixture_captured

    def disable_fixture(self):
        self._fixture_captured = None

    def activate_fixture(self):
        if self._fixture_captured is None:
            return

        assert self._manager is not None
        self._manager.destinations = [self._fixture_captured]

    def deactivate_fixture(self):
        assert self._manager is not None
        assert self._global_captured is not None
        self._manager.destinations = [self._global_captured]

    @contextlib.contextmanager
    def item_capture(self, when: str, item: pytest.Item) -> Generator[None]:
        self.start_global_capturing()
        self.activate_fixture()
        try:
            yield
        finally:
            self.deactivate_fixture()
            self.stop_global_capturing()

            out = self.read_global_capture()
            item.add_report_section(when, "loglady", out)

    # Hooks

    @pytest.hookimpl(wrapper=True)
    def pytest_runtest_setup(self, item: pytest.Item) -> Generator[None]:
        with self.item_capture("setup", item):
            return (yield)

    @pytest.hookimpl(wrapper=True)
    def pytest_runtest_call(self, item: pytest.Item) -> Generator[None]:
        with self.item_capture("call", item):
            return (yield)

    @pytest.hookimpl(wrapper=True)
    def pytest_runtest_teardown(self, item: pytest.Item) -> Generator[None]:
        with self.item_capture("teardown", item):
            return (yield)

    @pytest.hookimpl(tryfirst=True)
    def pytest_keyboard_interrupt(self) -> None:
        self.stop_global_capturing()

    @pytest.hookimpl(tryfirst=True)
    def pytest_internalerror(self) -> None:
        self.stop_global_capturing()


@pytest.fixture
def loglady_capture(request: pytest.FixtureRequest) -> Generator[list[Record], None, None]:
    """A fixture that captures global loglady logs and yields the list of captured logs"""
    plugin = request.config.pluginmanager.getplugin("loglady-plugin")
    assert isinstance(plugin, LogladyPlugin)

    capture = plugin.enable_fixture()
    yield capture.records
    plugin.disable_fixture()



# Copyright (c) 2024 Alethea Katherine Flowers.
# Published under the standard MIT License.
# Full text available at: https://opensource.org/licenses/MIT

"""Helpers for threading"""

import collections
import threading

# Note: this list is chosen based on how easy it is to tell the emoji apart and how well they display on a terminal.
_EMOJI = list(
    "👽🤖🎃🥶🦷👂👀👤🧶🧵🧦🧤🎩👑💍🌂🏀🏈🎾🎱🏓🪃🪁🏹🥊🛹🛼🥌🏆🥇🎪🎭🎨🎬🎤🎧🪇🥁🎷🎺🪗🎸🪕🎻🎲🎳🎮🧩🚗🛞🚀🛸🚁🛶🛟🗿🎡🎠💾💿📼📷🧭⏰📡💡🔦💎🪚🧲🔮💈🦠🧽🧸🎁🎈🎀🪭🪩📦📯📁📎🩷💮🌀📣🍏🍐🍊🍋🍌🍉🍇🫐🍈🍒🍑🥭🍍🥥🥝🍅🍆🥑🫛🥦🥬🫑🌽🥕🧄🧅🥐🍞🥖🥨🧀🥚🥓🥩🍖🦴🌭🍔🍟🍕🥪🌮🌯🥗🥫🍣🥟🦪🍤🍚🍥🥠🍡🍨🍦🥧🧁🍰🍮🍭🍬🍫🍿🍩🍪🌰🥜🍯🫖🧃🧊🥡🧂🐶🐱🐭🐹🐰🦊🐻🐼🐨🐯🦁🐮🐷🐽🐸🐵🐔🐧🐦🪿🦆🦉🦇🐺🐗🐴🦄🫎🐝🪱🐛🦋🐌🐞🐜🪲🦂🐢🐍🦎🦖🦕🐙🪼🦐🦞🦀🐡🐠🐟🐬🐳🦈🦭🐊🦍🦧🐩🐓🦃🦤🦚🦜🦢🦩🐇🦝🦨🦡🦫🦦🦥🐁🐀🦔🐉"
)

_thread_id_to_emoji = collections.OrderedDict()


def thread_emoji(thread_id: int | threading.Thread | None = None):
    """Cute little helper that assigns each thread a unique emoji."""
    if isinstance(thread_id, threading.Thread):
        thread_id = thread_id.ident

    if thread_id is None:
        thread_id = threading.get_ident()

    is_main_thread = thread_id == threading.main_thread().ident

    if is_main_thread:
        return "⭐️"

    # Note: this used to just use the thread_id to index into the _EMOJI list with modulo, however, it turns out that
    # thread ids can sometime be evenly spaced which ruins all the fun of modulo indexing, leading to only a very small
    # number of emojis being used. Instead, we do this nonsense.

    if thread_id in _thread_id_to_emoji:
        return _thread_id_to_emoji[thread_id]

    _thread_id_to_emoji[thread_id] = _EMOJI[len(_thread_id_to_emoji) % len(_EMOJI)]

    # This limits the size of the dictionary.
    while len(_thread_id_to_emoji) > len(_EMOJI):
        _ = _thread_id_to_emoji.popitem(last=False)

    return _thread_id_to_emoji[thread_id]



# Copyright (c) 2024 Alethea Katherine Flowers.
# Published under the standard MIT License.
# Full text available at: https://opensource.org/licenses/MIT

from __future__ import annotations

from collections.abc import Callable, Sequence
from types import TracebackType
from typing import Any

type Record = dict[str, Any]
type Context = Record
type Middleware = Callable[[Record], Record | None]
type MiddlewareList = Sequence[Middleware]
type Relay = Callable[[Record], None]
type SysExcInfo = tuple[type[BaseException], BaseException, TracebackType | None] | tuple[None, None, None]
